# Datasets

Currently the database is hosted on the following remote file share server: 
[{{ file_share_url }}]({{ file_share_url }}).

## Available Datasets

The following datasets can be downloaded either using the command line interface (CLI) or directly 
using the python API.

{{ datasets_table }}