[cache]
use_cache = true

[remote]
fileshare_type = "nextcloud"
fileshare_url = "https://bwsyncandshare.kit.edu/s/98NgmGCDty54kik/download/"

[remote.nextcloud]
a = 1
#dav_url = "https://nextcloud.DOMAIN.com/remote.php/dav/files/USERNAME/ChemMatData"
#dav_username = "USERNAME"
#dav_password = "PASSWORD"